program KeyPeek;
{ KeyPeek version 1.1.0 }
{ Copyright (c) 1999  Marius Gedminas <mgedmin@takas.lt> }

{ $Id: keypeek.pas,v 1.4 1999/08/25 20:38:52 mg Exp $ }

{ This program is free software; you can redistribute it and/or modify }
{ it under the terms of the GNU General Public License as published by }
{ the Free Software Foundation; either version 2 of the License, or    }
{ (at your option) any later version.                                  }
{                                                                      }
{ This program is distributed in the hope that it will be useful,      }
{ but WITHOUT ANY WARRANTY; without even the implied warranty of       }
{ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        }
{ GNU General Public License for more details.                         }
{                                                                      }
{ You should have received a copy of the GNU General Public License    }
{ along with this program; if not, write to the Free Software          }
{ Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            }

{ History: }

{ 1996-07-02  Version 1.00 of KEYPEEK.ASM }
{   The last modification date of that version.                        }

{ 1999-02-20  Version 1.01 of KEYPEEK.ASM }
{   Added many comments, placed under CVS source control and GPL'ed.   }

{ 1999-02-20  Version 1.2 }
{   Rewritten in Borland Pascal 7.0                                     }
{   Tries to detect keyboard functionality                              }
{   Displays key names suitable for PutKb                               }

const
  PrgName = 'KeyPeek';
  Version = '1.2';

const
  ShiftPrefix = 'Shift-';
  CtrlPrefix = 'Ctrl-';
  AltPrefix = 'Alt-';

const
  kpNone  = 0;
  kpCtrl  = 1;
  kpAlt   = 2;
  kpShift = 3;
  Prefixes: array[kpNone..kpShift] of string[6] = (
    '',
    CtrlPrefix,
    AltPrefix,
    ShiftPrefix);

const
  { These come straight from PutKb.pas, so are a bit redundant }
  NumKeys = 59;
  KeyNames: array[1..NumKeys] of string[11] = (
    'Esc', 'Escape',
    'F1',
    'F2',
    'F3',
    'F4',
    'F5',
    'F6',
    'F7',
    'F8',
    'F9',
    'F10',
    'F11',
    'F12',
    'Back', 'Backspace', 'Bksp',
    'Tab',
    'Enter', 'Return',
    'Space',
    'Home',
    'Up',
    'PgUp', 'PageUp',
    'Left',
    'Center', 'Key5',
    'Right',
    'End',
    'Down',
    'PgDn', 'PageDown',
    'Ins', 'Insert',
    'Del', 'Delete',
    'NumEnter',
    'Plus',
    'Minus',
    'Star', 'Asterisk',
    'Slash',
    'GIns', 'GInsert',
    'GDel', 'GDelete',
    'GHome',
    'GEnd',
    'GPgUp', 'GPageUp',
    'GPgDn', 'GPageDown',
    'GUp',
    'GDown',
    'GLeft',
    'GRight',
    'PrtScr', 'PrintScreen');

  { Values are those that appear in keyboard buffer.  INT 16h functions }
  { return slightly different values (e.g. Center: 4CF0h and 4C00h) }
  KeyCodes: array[kpNone..kpShift, 1..NumKeys] of Word = (
  ({ kpNone }
    $011B, $011B,                 { 'Esc', 'Escape', }
    $3B00,                        { 'F1', }
    $3C00,                        { 'F2', }
    $3D00,                        { 'F3', }
    $3E00,                        { 'F4', }
    $3F00,                        { 'F5', }
    $4000,                        { 'F6', }
    $4100,                        { 'F7', }
    $4200,                        { 'F8', }
    $4300,                        { 'F9', }
    $4400,                        { 'F10', }
    $8500,                        { 'F11', }
    $8600,                        { 'F12', }
    $0E08, $0E08, $0E08,          { 'Back', 'Backspace', 'Bksp', }
    $0F09,                        { 'Tab', }
    $1C0D, $1C0D,                 { 'Enter', 'Return', }
    $3920,                        { 'Space', }
    $4700,                        { 'Home', }
    $4800,                        { 'Up', }
    $4900, $4900,                 { 'PgUp', 'PageUp', }
    $4B00,                        { 'Left', }
    $4CF0, $4CF0,                 { 'Center', 'Key5', }
    $4D00,                        { 'Right', }
    $4F00,                        { 'End', }
    $5000,                        { 'Down', }
    $5100, $5100,                 { 'PgDn', 'PageDown', }
    $5200, $5200,                 { 'Ins', 'Insert', }
    $5300, $5300,                 { 'Del', 'Delete', }
    $E00D,                        { 'NumEnter', }
    $4E2B,                        { 'Plus', }
    $4A2D,                        { 'Minus', }
    $372A, $372A,                 { 'Star', 'Asterisk', }
    $E02F,                        { 'Slash', }
    $52E0, $52E0,                 { 'GIns', 'GInsert', }
    $53E0, $53E0,                 { 'GDel', 'GDelete', }
    $47E0,                        { 'GHome', }
    $4FE0,                        { 'GEnd', }
    $49E0, $49E0,                 { 'GPgUp', 'GPageUp', }
    $51E0, $51E0,                 { 'GPgDn', 'GPageDown', }
    $48E0,                        { 'GUp', }
    $50E0,                        { 'GDown', }
    $4BE0,                        { 'GLeft', }
    $4DE0,                        { 'GRight', }
        0,     0),                { 'PrtScr', 'PrintScreen'); }
  ({ kpCtrl }
    $011B, $011B,                 { 'Esc', 'Escape', }
    $5E00,                        { 'F1', }
    $5F00,                        { 'F2', }
    $6000,                        { 'F3', }
    $6100,                        { 'F4', }
    $6200,                        { 'F5', }
    $6300,                        { 'F6', }
    $6400,                        { 'F7', }
    $6500,                        { 'F8', }
    $6600,                        { 'F9', }
    $6700,                        { 'F10', }
    $8900,                        { 'F11', }
    $8A00,                        { 'F12', }
    $0E7F, $0E7F, $0E7F,          { 'Back', 'Backspace', 'Bksp', }
    $9400,                        { 'Tab', }
    $1C0A, $1C0A,                 { 'Enter', 'Return', }
    $3920,                        { 'Space', }
    $7700,                        { 'Home', }
    $8D00,                        { 'Up', }
    $8400, $8400,                 { 'PgUp', 'PageUp', }
    $7300,                        { 'Left', }
    $8F00, $8F00,                 { 'Center', 'Key5', }
    $7400,                        { 'Right', }
    $7500,                        { 'End', }
    $9100,                        { 'Down', }
    $7600, $7600,                 { 'PgDn', 'PageDown', }
    $9200, $9200,                 { 'Ins', 'Insert', }
    $9300, $9300,                 { 'Del', 'Delete', }
    $E00A,                        { 'NumEnter', }
    $9000,                        { 'Plus', }
    $8E00,                        { 'Minus', }
    $9600, $9600,                 { 'Star', 'Asterisk', }
    $9500,                        { 'Slash', }
    $92E0, $92E0,                 { 'GIns', 'GInsert', }
    $93E0, $93E0,                 { 'GDel', 'GDelete', }
    $77E0,                        { 'GHome', }
    $75E0,                        { 'GEnd', }
    $84E0, $84E0,                 { 'GPgUp', 'GPageUp', }
    $76E0, $76E0,                 { 'GPgDn', 'GPageDown', }
    $8DE0,                        { 'GUp', }
    $91E0,                        { 'GDown', }
    $73E0,                        { 'GLeft', }
    $74E0,                        { 'GRight', }
    $7200, $7200),                { 'PrtScr', 'PrintScreen'); }
  ({ kpAlt }
    $01F0, $01F0,                 { 'Esc', 'Escape', }
    $6800,                        { 'F1', }
    $6900,                        { 'F2', }
    $6A00,                        { 'F3', }
    $6B00,                        { 'F4', }
    $6C00,                        { 'F5', }
    $6D00,                        { 'F6', }
    $6E00,                        { 'F7', }
    $6F00,                        { 'F8', }
    $7000,                        { 'F9', }
    $7100,                        { 'F10', }
    $8B00,                        { 'F11', }
    $8C00,                        { 'F12', }
    $0EF0, $0EF0, $0EF0,          { 'Back', 'Backspace', 'Bksp', }
    $A500,                        { 'Tab', }
    $1CF0, $1CF0,                 { 'Enter', 'Return', }
    $3920,                        { 'Space', }
        0,                        { 'Home', }
        0,                        { 'Up', }
        0,     0,                 { 'PgUp', 'PageUp', }
        0,                        { 'Left', }
        0,     0,                 { 'Center', 'Key5', }
        0,                        { 'Right', }
        0,                        { 'End', }
        0,                        { 'Down', }
        0,     0,                 { 'PgDn', 'PageDown', }
        0,     0,                 { 'Ins', 'Insert', }
        0,     0,                 { 'Del', 'Delete', }
    $A600,                        { 'NumEnter', }
    $4EF0,                        { 'Plus', }
    $4AF0,                        { 'Minus', }
    $37F0, $37F0,                 { 'Star', 'Asterisk', }
    $A400,                        { 'Slash', }
    $A200, $A200,                 { 'GIns', 'GInsert', }
    $A300, $A300,                 { 'GDel', 'GDelete', }
    $9700,                        { 'GHome', }
    $9F00,                        { 'GEnd', }
    $9900, $9900,                 { 'GPgUp', 'GPageUp', }
    $A100, $A100,                 { 'GPgDn', 'GPageDown', }
    $9800,                        { 'GUp', }
    $A000,                        { 'GDown', }
    $9B00,                        { 'GLeft', }
    $9D00,                        { 'GRight', }
        0,     0),                { 'PrtScr', 'PrintScreen'); }
  ({ kpShift }
    $011B, $011B,                 { 'Esc', 'Escape', }
    $5400,                        { 'F1', }
    $5500,                        { 'F2', }
    $5500,                        { 'F3', }
    $5700,                        { 'F4', }
    $5800,                        { 'F5', }
    $5900,                        { 'F6', }
    $5A00,                        { 'F7', }
    $5B00,                        { 'F8', }
    $5C00,                        { 'F9', }
    $5D00,                        { 'F10', }
    $8700,                        { 'F11', }
    $8800,                        { 'F12', }
    $0E08, $0E08, $0E08,          { 'Back', 'Backspace', 'Bksp', }
    $0F00,                        { 'Tab', }
    $1C0D, $1C0D,                 { 'Enter', 'Return', }
    $3920,                        { 'Space', }
    $4737,                        { 'Home', }
    $4838,                        { 'Up', }
    $4939, $4939,                 { 'PgUp', 'PageUp', }
    $4B34,                        { 'Left', }
    $4C35, $4C35,                 { 'Center', 'Key5', }
    $4D36,                        { 'Right', }
    $4F31,                        { 'End', }
    $5032,                        { 'Down', }
    $5133, $5133,                 { 'PgDn', 'PageDown', }
    $5230, $5230,                 { 'Ins', 'Insert', }
    $532E, $532E,                 { 'Del', 'Delete', }
    $E00D,                        { 'NumEnter', }
    $4E2B,                        { 'Plus', }
    $4A2D,                        { 'Minus', }
    $372A, $372A,                 { 'Star', 'Asterisk', }
    $E02F,                        { 'Slash', }
    $52E0, $52E0,                 { 'GIns', 'GInsert', }
    $53E0, $53E0,                 { 'GDel', 'GDelete', }
    $47E0,                        { 'GHome', }
    $4FE0,                        { 'GEnd', }
    $49E0, $49E0,                 { 'GPgUp', 'GPageUp', }
    $51E0, $51E0,                 { 'GPgDn', 'GPageDown', }
    $48E0,                        { 'GUp', }
    $50E0,                        { 'GDown', }
    $4BE0,                        { 'GLeft', }
    $4DE0,                        { 'GRight', }
        0,     0));               { 'PrtScr', 'PrintScreen'); }

  NumChars = 95;
  Chars: array[1..NumChars] of Char =
    ' '+
    '`~'+
    '1!'+
    '2@'+
    '3#'+
    '4$'+
    '5%'+
    '6^'+
    '7&'+
    '8*'+
    '9('+
    '0)'+
    '-_'+
    '=+'+
    '\|'+
    '[{'+
    ']}'+
    ';:'+
    '''"'+
    ',<'+
    '.>'+
    '/?'+
    'qQ'+
    'wW'+
    'eE'+
    'rR'+
    'tT'+
    'yY'+
    'uU'+
    'iI'+
    'oO'+
    'pP'+
    'aA'+
    'sS'+
    'dD'+
    'fF'+
    'gG'+
    'hH'+
    'jJ'+
    'kK'+
    'lL'+
    'zZ'+
    'xX'+
    'cC'+
    'vV'+
    'bB'+
    'nN'+
    'mM';

  CharCodes: array[kpNone..kpAlt, 1..NumChars] of Word = (
  ({ kpNone }
    $3920,                   { ' '+ }
    $2960, $297E,            { '`~'+ }
    $0231, $0221,            { '1!'+ }
    $0332, $0340,            { '2@'+ }
    $0433, $0423,            { '3#'+ }
    $0534, $0524,            { '4$'+ }
    $0635, $0625,            { '5%'+ }
    $0736, $075E,            { '6^'+ }
    $0837, $0826,            { '7&'+ }
    $0938, $092A,            { '8*'+ }
    $0A39, $0A28,            { '9('+ }
    $0B30, $0B29,            { '0)'+ }
    $0C2D, $0C5F,            { '-_'+ }
    $0D3D, $0D2B,            { '=+'+ }
    $2B5C, $2B7C,            { '\|'+ }
    $1A5B, $1A7B,            { '[{'+ }
    $1B5D, $1B7D,            (*']}'+*)
    $273B, $273A,            { ';:'+ }
    $2827, $2822,            { '''"'+ }
    $332C, $333C,            { ',<'+ }
    $342E, $343E,            { '.>'+ }
    $352F, $353F,            { '/?'+ }
    $1071, $1051,            { 'qQ'+ }
    $1177, $1157,            { 'wW'+ }
    $1265, $1245,            { 'eE'+ }
    $1372, $1352,            { 'rR'+ }
    $1474, $1454,            { 'tT'+ }
    $1579, $1559,            { 'yY'+ }
    $1675, $1655,            { 'uU'+ }
    $1769, $1749,            { 'iI'+ }
    $186F, $184F,            { 'oO'+ }
    $1970, $1950,            { 'pP'+ }
    $1E61, $1E41,            { 'aA'+ }
    $1F73, $1F53,            { 'sS'+ }
    $2064, $2044,            { 'dD'+ }
    $2166, $2146,            { 'fF'+ }
    $2267, $2247,            { 'gG'+ }
    $2368, $2348,            { 'hH'+ }
    $246A, $244A,            { 'jJ'+ }
    $256B, $254B,            { 'kK'+ }
    $266C, $264C,            { 'lL'+ }
    $2C7A, $48E0,            { 'zZ'+ }
    $2D78, $2D58,            { 'xX'+ }
    $2E63, $2E43,            { 'cC'+ }
    $2F76, $2F56,            { 'vV'+ }
    $3062, $3042,            { 'bB'+ }
    $316E, $314E,            { 'nN'+ }
    $326D, $324D),           { 'mM'; }
  ({ kpCtrl }
    $3920,                   { ' '+ }
        0,     0,            { '`~'+ }
        0,     0,            { '1!'+ }
    $0300, $0300,            { '2@'+ }
        0,     0,            { '3#'+ }
        0,     0,            { '4$'+ }
        0,     0,            { '5%'+ }
    $071E, $071E,            { '6^'+ }
        0,     0,            { '7&'+ }
        0,     0,            { '8*'+ }
        0,     0,            { '9('+ }
        0,     0,            { '0)'+ }
    $0C1F, $0C1F,            { '-_'+ }
        0,     0,            { '=+'+ }
    $2B1C, $2B1C,            { '\|'+ }
    $1A1B, $1A1B,            { '[{'+ }
    $1B1D, $1B1D,            (*']}'+*)
        0,     0,            { ';:'+ }
        0,     0,            { '''"'+ }
        0,     0,            { ',<'+ }
        0,     0,            { '.>'+ }
        0,     0,            { '/?'+ }
    $1011, $1011,            { 'qQ'+ }
    $1117, $1117,            { 'wW'+ }
    $1205, $1205,            { 'eE'+ }
    $1312, $1312,            { 'rR'+ }
    $1414, $1414,            { 'tT'+ }
    $1519, $1519,            { 'yY'+ }
    $1615, $1615,            { 'uU'+ }
    $1709, $1709,            { 'iI'+ }
    $180F, $180F,            { 'oO'+ }
    $1910, $1910,            { 'pP'+ }
    $1E01, $1E01,            { 'aA'+ }
    $1F13, $1F13,            { 'sS'+ }
    $2004, $2004,            { 'dD'+ }
    $2106, $2106,            { 'fF'+ }
    $2207, $2207,            { 'gG'+ }
    $2308, $2308,            { 'hH'+ }
    $240A, $240A,            { 'jJ'+ }
    $250B, $250B,            { 'kK'+ }
    $260C, $260C,            { 'lL'+ }
    $2C1A, $2C1A,            { 'zZ'+ }
    $2D18, $2D18,            { 'xX'+ }
    $2E03, $2E03,            { 'cC'+ }
    $2F16, $2F16,            { 'vV'+ }
    $3002, $3002,            { 'bB'+ }
    $310E, $310E,            { 'nN'+ }
    $320D, $320D),           { 'mM';  }
  ({ kpAlt }
    $3920,                   { ' '+ }
    $29F0, $29F0,            { '`~'+ }
    $7800, $7800,            { '1!'+ }
    $7900, $7900,            { '2@'+ }
    $7A00, $7A00,            { '3#'+ }
    $7B00, $7B00,            { '4$'+ }
    $7C00, $7C00,            { '5%'+ }
    $7D00, $7D00,            { '6^'+ }
    $7E00, $7E00,            { '7&'+ }
    $7F00, $7F00,            { '8*'+ }
    $8000, $8000,            { '9('+ }
    $8100, $8100,            { '0)'+ }
    $8200, $8200,            { '-_'+ }
    $8300, $8300,            { '=+'+ }
    $2BF0, $2BF0,            { '\|'+ }
    $1AF0, $1AF0,            { '[{'+ }
    $1BF0, $1BF0,            (*']}'+*)
    $27F0, $27F0,            { ';:'+ }
    $28F0, $28F0,            { '''"'+ }
    $33F0, $33F0,            { ',<'+ }
    $34F0, $34F0,            { '.>'+ }
    $35F0, $35F0,            { '/?'+ }
    $1000, $1000,            { 'qQ'+ }
    $1100, $1100,            { 'wW'+ }
    $1200, $1200,            { 'eE'+ }
    $1300, $1300,            { 'rR'+ }
    $1400, $1400,            { 'tT'+ }
    $1500, $1500,            { 'yY'+ }
    $1600, $1600,            { 'uU'+ }
    $1700, $1700,            { 'iI'+ }
    $1800, $1800,            { 'oO'+ }
    $1900, $1900,            { 'pP'+ }
    $1E00, $1E00,            { 'aA'+ }
    $1F00, $1F00,            { 'sS'+ }
    $2000, $2000,            { 'dD'+ }
    $2100, $2100,            { 'fF'+ }
    $2200, $2200,            { 'gG'+ }
    $2300, $2300,            { 'hH'+ }
    $2400, $2400,            { 'jJ'+ }
    $2500, $2500,            { 'kK'+ }
    $2600, $2600,            { 'lL'+ }
    $2C00, $2C00,            { 'zZ'+ }
    $2D00, $2D00,            { 'xX'+ }
    $2E00, $2E00,            { 'cC'+ }
    $2F00, $2F00,            { 'vV'+ }
    $3000, $3000,            { 'bB'+ }
    $3100, $3100,            { 'nN'+ }
    $3200, $3200));          { 'mM'; }

procedure PrintKey(Code: Word);
  function DoPrint(Code: Word): Boolean;
  var
    I, J: Integer;
  begin
    DoPrint := True;
    for J := kpNone to kpShift do
      for I := 1 to NumKeys do
        if Code = KeyCodes[J, I] then
          begin
            Write(Prefixes[J], KeyNames[I]);
            Exit;
          end;
    for J := kpNone to kpAlt do
      for I := 1 to NumChars do
        if Code = CharCodes[J, I] then
          begin
            if Chars[I] = '''' then Write(Prefixes[J], '"''"')
            else Write(Prefixes[J], '''', Chars[I], '''');
            Exit;
          end;
    if (Hi(Code) = 0) and (Lo(Code) >= $20) then
      begin { Alt-NNN }
        if Chr(Code) = '''' then Write('"''"')
        else Write('''', Chr(Code), '''');
        Exit;
      end;
    DoPrint := False;
  end;
begin
  if Code = 0 then
    begin
      { Ctrl-Break places this on my system }
      Exit;
    end;
  if DoPrint(Code) then
     Exit;
  if (Lo(Code) >= $20) then
    begin { Non standard }
      if Chr(Code) = '''' then Write('("''")')
      else Write('(''', Chr(Code), ''')');
      Exit;
    end;
  if (Lo(Code) = $00) then
    begin { Maybe someone placed it directly }
      if DoPrint(Code + $F0) then
        Exit;
    end;
end;


type
  Str8 = string[8];

const
  HexChars: array[0..$F] of Char = '0123456789ABCDEF';

function HexWord(N: Word): Str8;
begin
  HexWord[0] := Char(4);
  HexWord[1] := HexChars[(N shr 12)       ];
  HexWord[2] := HexChars[(N shr  8) and $F];
  HexWord[3] := HexChars[(N shr  4) and $F];
  HexWord[4] := HexChars[(N       ) and $F];
end;

const
  HaveExtKeyb: Boolean = False;         { 101/102 key keyboard }
  HaveExt2Keyb: Boolean = False;        { 122 key keyboard }

procedure DetectKeyb;
type
  PtrRec = record
    Ofs, Seg: Word;
  end;
  PInfoTable = ^TInfoTable;
  TInfoTable = record
    Size: Word;
    Model: Byte;
    Submodel: Byte;
    Revision: Byte;
    Feature1: Byte;
    Feature2: Byte;
    { and so on }
  end;
var
  TablePtr: PInfoTable;
  Res: Byte;
begin
  HaveExtKeyb := False;
  HaveExt2Keyb := False;

  if Mem[Seg0040:$96] and (1 shl 4) <> 0 then
    HaveExtKeyb := True;

  asm
    mov  ah, 0C0h
    int  15h
    { INT 15 fn 0C0h - Get configuration }
    { Available on: XT after 1/10/86,AT mdl 3x9,CONV,XT286,PS }
    jnc  @@1
    { Carry set -- function not supported by BIOS }
    xor  bx, bx
    mov  es, bx
  @@1:
    mov  TablePtr.PtrRec.&Seg, es
    mov  TablePtr.PtrRec.&Ofs, bx
  end;
  if TablePtr = nil then
    Exit; { ROM configuration table not available }
  if TablePtr^.Size < SizeOf(TablePtr^) then
    Exit; { Table does not have Feature2 -- shouldn't happen }
  if TablePtr^.Feature2 and (1 shl 6) = 0 then
    Exit; { INT 16h/AH=09h (keyboard functionality) not supported }
  asm
    mov  ax, 0900h
    int  16h
    { INF 16h fn 09h - Get keyboard functionality }
    mov  Res, al
  end;
  HaveExtKeyb := Res and (1 shl 5) <> 0;
  HaveExt2Keyb := Res and (1 shl 6) <> 0;
end;

procedure ClearKeyb;
begin
  MemW[Seg0040:$1A] := MemW[Seg0040:$1C];
end;

function PeekBuf(var Key: Word): Boolean;
begin
  if MemW[Seg0040:$1A] = MemW[Seg0040:$1C] then
    PeekBuf := False
  else
    begin
      PeekBuf := True;
      Key := MemW[Seg0040:MemW[Seg0040:$1A]];
    end;
end;

function PeekStd(var Key: Word): Boolean; assembler;
asm
  mov  ah, 1
  int  16h
  mov  cx, False
  jz   @@1
  les  bx, Key
  mov  es:[bx], ax
  inc  cx { True }
@@1:
  mov  ax, cx
end;

function PeekExt(var Key: Word): Boolean; assembler;
asm
  mov  ah, 11h
  int  16h
  mov  cx, False
  jz   @@1
  les  bx, Key
  mov  es:[bx], ax
  inc  cx { True }
@@1:
  mov  ax, cx
end;

function PeekExt2(var Key: Word): Boolean; assembler;
asm
  mov  ah, 21h
  int  16h
  mov  cx, False
  jz   @@1
  les  bx, Key
  mov  es:[bx], ax
  inc  cx { True }
@@1:
  mov  ax, cx
end;

var
  Std, Ext, Ext2, Buf: Word;
const
  HaveStd: Boolean = False;
  HaveExt: Boolean = False;
  HaveExt2: Boolean = False;

begin
  WriteLn(PrgName+' v'+Version+'  Copyright (c) 1999, M. Gedminas <mgedmin@takas.lt>');
  WriteLn('This program is Free Software.  See file COPYING for details.');
  WriteLn;
  DetectKeyb;
  if HaveExt2Keyb then WriteLn('122-key keyboard supported by BIOS')
  else if HaveExtKeyb then WriteLn('Enhanced (101/102-key) keyboard supported by BIOS');
  Write('Std ');
  if HaveExtKeyb then Write(' Ext ');
  if HaveExt2Keyb then Write(' Ext2');
  WriteLn(' Buf');
  Write('----');
  if HaveExtKeyb then Write(' ----');
  if HaveExt2Keyb then Write(' ----');
  WriteLn(' ----');
  ClearKeyb;
  repeat
    repeat
    until PeekBuf(Buf);
    if HaveExt2Keyb then
      HaveExt2 := PeekExt2(Ext2);
    if HaveExtKeyb then
      HaveExt := PeekExt(Ext);
    HaveStd := PeekStd(Std);
    ClearKeyb;
    if HaveStd then
      Write(HexWord(Std), ' ')
    else
      Write('---- ');
    if HaveExtKeyb then
      if HaveExt then
        Write(HexWord(Ext), ' ')
      else
        Write('---- ');
    if HaveExt2Keyb then
      if HaveExt2 then
        Write(HexWord(Ext2), ' ')
      else
        Write('---- ');
    Write(HexWord(Buf), '  ');
    PrintKey(Buf);
    WriteLn;
  until Buf = $011B; { Esc }
end.

{ -------------------------------------------------------------- The End - }
